@@ -5,6 +5,9 @@ use std::mem::swap;
 extern crate test;
 use test::{black_box, Bencher};
 
+#[cfg(feature = "rayon")]
+use rayon::iter::{IndexedParallelIterator, ParallelIterator};
+
 use rs_ecs::{Entity, Query, Resources, World};
 
 struct Pos(f32);
@@ -119,6 +122,31 @@ fn insert_remove(bencher: &mut Bencher) {
     });
 }
 
+#[bench]
+fn exchange(bencher: &mut Bencher) {
+    let mut world = World::new();
+
+    spawn_few(&mut world);
+
+    let entities = Query::<&Entity>::new()
+        .borrow(&world)
+        .iter()
+        .copied()
+        .collect::<Vec<_>>();
+
+    let mut entities = entities.iter().cycle();
+
+    bencher.iter(|| {
+        let world = black_box(&mut world);
+        let entities = black_box(&mut entities);
+
+        let ent = *entities.next().unwrap();
+
+        world.exchange::<(Pos,), _>(ent, (Pos(0.0),));
+        world.exchange::<(Vel,), _>(ent, (Vel(0.0),));
+    });
+}
+
 #[bench]
 fn transfer(bencher: &mut Bencher) {
     let mut world = World::new();
@@ -263,6 +291,39 @@ fn query_map_very_many_small_archetypes(bencher: &mut Bencher) {
     query_map(bencher, spawn_few_in_very_many_small_archetypes);
 }
 
+fn query_for_each(bencher: &mut Bencher, spawn: fn(&mut World)) {
+    let mut world = World::new();
+    let mut query = Query::<(&mut Pos, &Vel)>::new();
+
+    spawn(&mut world);
+
+    let _ = query.borrow(&world).iter();
+
+    bencher.iter(|| {
+        let world = black_box(&world);
+        let query = black_box(&mut query);
+
+        query.borrow(world).for_each(|(pos, vel)| {
+            pos.0 += vel.0;
+        });
+    });
+}
+
+#[bench]
+fn query_for_each_single_archetype(bencher: &mut Bencher) {
+    query_for_each(bencher, spawn_few);
+}
+
+#[bench]
+fn query_for_each_many_archetypes(bencher: &mut Bencher) {
+    query_for_each(bencher, spawn_few_in_many_archetypes);
+}
+
+#[bench]
+fn query_for_each_very_many_small_archetypes(bencher: &mut Bencher) {
+    query_for_each(bencher, spawn_few_in_very_many_small_archetypes);
+}
+
 #[bench]
 fn get_resource(bencher: &mut Bencher) {
     let mut resources = Resources::new();
@@ -275,3 +336,44 @@ fn get_resource(bencher: &mut Bencher) {
         let _f64 = resources.get::<f64>();
     });
 }
+
+#[cfg(feature = "rayon")]
+fn query_parallel(bencher: &mut Bencher, spawn: fn(&mut World)) {
+    let mut world = World::new();
+    let mut query = Query::<(&mut Pos, &Vel)>::new();
+
+    spawn(&mut world);
+
+    let _ = query.borrow(&world).iter();
+
+    bencher.iter(|| {
+        let world = black_box(&world);
+        let query = black_box(&mut query);
+
+        query
+            .borrow(world)
+            .par_iter()
+            .with_min_len(1024)
+            .for_each(|(pos, vel)| {
+                pos.0 += vel.0;
+            });
+    });
+}
+
+#[cfg(feature = "rayon")]
+#[bench]
+fn query_parallel_single_archetype(bencher: &mut Bencher) {
+    query_parallel(bencher, spawn_few);
+}
+
+#[cfg(feature = "rayon")]
+#[bench]
+fn query_parallel_many_archetypes(bencher: &mut Bencher) {
+    query_parallel(bencher, spawn_few_in_many_archetypes);
+}
+
+#[cfg(feature = "rayon")]
+#[bench]
+fn query_parallel_very_many_small_archetypes(bencher: &mut Bencher) {
+    query_parallel(bencher, spawn_few_in_very_many_small_archetypes);
+}
