@@ -11,6 +11,9 @@ use crate::{
     world::{Entity, EntityMetadata, World},
 };
 
+#[cfg(feature = "rayon")]
+use crate::rayon::QueryParIter;
+
 /// Query to get an iterator over all entities with a certain combination of components.
 ///
 /// Queries are specified through their type argument, by composing the type of their result.
@@ -240,6 +243,26 @@ impl<S> QueryRef<'_, S>
 where
     S: QuerySpec,
 {
+    /// Visit all entities matching the query.
+    pub fn for_each<'q, F>(&'q mut self, mut f: F)
+    where
+        F: FnMut(<S::Fetch as Fetch<'q>>::Item),
+    {
+        let types: &'q [(u16, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
+
+        for (idx, ty) in types {
+            let archetype = &self.world.archetypes[*idx as usize];
+
+            let ptr = unsafe { S::Fetch::base_pointer(archetype, *ty) };
+
+            for idx in 0..archetype.len() {
+                let val = unsafe { S::Fetch::deref(ptr, idx) };
+
+                f(val);
+            }
+        }
+    }
+
     /// Create an iterator over the entities matching the query.
     pub fn iter<'q>(&'q mut self) -> QueryIter<'q, S> {
         let types: &'q [(u16, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
@@ -253,6 +276,18 @@ where
         }
     }
 
+    #[cfg(feature = "rayon")]
+    /// Create a parallel iterator over the entities matching the query.
+    pub fn par_iter<'q>(&'q mut self) -> QueryParIter<'q, S>
+    where
+        <S::Fetch as Fetch<'q>>::Ty: Send + Sync,
+        <S::Fetch as Fetch<'q>>::Item: Send,
+    {
+        let types: &'q [(u16, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
+
+        QueryParIter::new(types, &self.world.archetypes)
+    }
+
     /// Create a map of the entities matching the query.
     ///
     /// This is an alternative to [get](World::get) and [get_mut](World::get_mut) for repeated calls, to amortize the set-up costs.
@@ -277,6 +312,7 @@ where
     /// ```
     pub fn map<'q>(&'q mut self) -> QueryMap<'q, S> {
         let types: &'q [(u16, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
+
         let ptrs: &'q mut Vec<Option<<S::Fetch as Fetch<'q>>::Ptr>> =
             unsafe { transmute(&mut *self.ptrs) };
 
@@ -396,6 +432,7 @@ pub trait QuerySpec {
     type Fetch: for<'a> Fetch<'a>;
 }
 
+#[allow(clippy::missing_safety_doc)]
 pub unsafe trait Fetch<'q> {
     type Ty: Copy;
     type Ref;
