@@ -15,6 +15,7 @@ pub struct World {
     pub(crate) archetypes: Vec<Archetype>,
     insert_map: IndexTypeIdMap<u16>,
     remove_map: IndexTypeIdMap<u16>,
+    exchange_map: IndexTypeIdMap<u16>,
     transfer_map: IndexTagMap<u16>,
 }
 
@@ -38,6 +39,7 @@ impl World {
             archetypes: vec![Archetype::new(empty_archetype)],
             insert_map: Default::default(),
             remove_map: Default::default(),
+            exchange_map: Default::default(),
             transfer_map: Default::default(),
         }
     }
@@ -113,19 +115,22 @@ impl World {
 
         meta.bump_gen();
 
-        Self::free_idx(
+        Self::free_idx::<true>(
             &mut self.archetypes[meta.ty as usize],
             meta.idx,
-            true,
             &mut self.entities,
         );
 
         self.free_list.push(ent.id);
     }
 
-    fn free_idx(archetype: &mut Archetype, idx: u32, drop: bool, entities: &mut [EntityMetadata]) {
+    fn free_idx<const DROP: bool>(
+        archetype: &mut Archetype,
+        idx: u32,
+        entities: &mut [EntityMetadata],
+    ) {
         unsafe {
-            if archetype.free(idx, drop) {
+            if archetype.free::<DROP>(idx) {
                 let swapped_ent = archetype.pointer::<Entity>(idx).read();
 
                 entities[swapped_ent.id as usize].idx = idx;
@@ -284,6 +289,81 @@ impl World {
         Some(new_ty)
     }
 
+    /// Exchange components for a given [Entity]
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// # use rs_ecs::*;
+    /// let mut world = World::new();
+    ///
+    /// let entity = world.alloc();
+    /// world.insert(entity, (42_u32, true));
+    /// assert!(world.contains::<u32>(entity));
+    /// assert!(world.contains::<bool>(entity));
+    /// assert!(!world.contains::<String>(entity));
+    ///
+    /// world.exchange::<(u32, bool), _>(entity, (String::from("Hello"),)).unwrap();
+    /// assert!(!world.contains::<u32>(entity));
+    /// assert!(!world.contains::<bool>(entity));
+    /// assert!(world.contains::<String>(entity));
+    /// ```
+    pub fn exchange<B1, B2>(&mut self, ent: Entity, new_comps: B2) -> Option<B1>
+    where
+        B1: Bundle,
+        B2: Bundle,
+    {
+        let meta = &self.entities[ent.id as usize];
+        assert_eq!(ent.gen, meta.gen, "Entity is stale");
+
+        let new_ty = if let Some(ty) = self.exchange_map.get(&(meta.ty, TypeId::of::<(B1, B2)>())) {
+            *ty
+        } else {
+            Self::exchange_cold(
+                &mut self.archetypes,
+                &mut self.exchange_map,
+                TypeId::of::<(B1, B2)>(),
+                B1::remove,
+                B2::insert,
+                meta.ty,
+            )?
+        };
+
+        let old_ty = meta.ty;
+        let old_idx = meta.idx;
+
+        unsafe {
+            let old_comps = B1::read(&mut self.archetypes[old_ty as usize], old_idx);
+
+            let new_idx = self.move_(ent.id, old_ty, new_ty, old_idx);
+
+            new_comps.write(&mut self.archetypes[new_ty as usize], new_idx);
+
+            Some(old_comps)
+        }
+    }
+
+    #[cold]
+    #[inline(never)]
+    fn exchange_cold(
+        archetypes: &mut Vec<Archetype>,
+        exchange_map: &mut IndexTypeIdMap<u16>,
+        type_id: TypeId,
+        remove: fn(&mut TypeMetadataSet) -> Option<()>,
+        insert: fn(&mut TypeMetadataSet),
+        old_ty: u16,
+    ) -> Option<u16> {
+        let mut types = archetypes[old_ty as usize].types();
+        remove(&mut types)?;
+        insert(&mut types);
+
+        let new_ty = Self::get_or_insert(archetypes, types);
+
+        exchange_map.insert((old_ty, type_id), new_ty);
+
+        Some(new_ty)
+    }
+
     fn get_or_insert(archetypes: &mut Vec<Archetype>, types: TypeMetadataSet) -> u16 {
         let pos = archetypes
             .iter_mut()
@@ -309,7 +389,7 @@ impl World {
 
         Archetype::move_(old_archetype, new_archetype, old_idx, new_idx);
 
-        Self::free_idx(old_archetype, old_idx, false, &mut self.entities);
+        Self::free_idx::<false>(old_archetype, old_idx, &mut self.entities);
 
         let meta = &mut self.entities[id as usize];
         meta.ty = new_ty;
@@ -375,7 +455,7 @@ impl World {
 
             Archetype::move_(old_archetype, new_archetype, meta.idx, new_meta.idx);
 
-            Self::free_idx(old_archetype, meta.idx, false, &mut self.entities);
+            Self::free_idx::<false>(old_archetype, meta.idx, &mut self.entities);
         }
 
         // fix entity component in other
@@ -545,6 +625,7 @@ impl EntityMetadata {
     }
 }
 
+#[allow(clippy::missing_safety_doc)]
 pub unsafe trait Bundle: 'static {
     fn insert(types: &mut TypeMetadataSet);
     #[must_use]
